/*
    sud0 - run preconfigured commands as root or another user
    Copyright (C) 2022  Tibor 'Igor2' Palinkas

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Source: svn://svn.repo.hu/sud0/trunk
    Contact: http://igor2.repo.hu/contact.html
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

#include "qparse.c"
#include "config.h"
#include "version.h"

#define NL "\n"

#define HELP \
	"sud0 - run a command line as another user" NL NL\
	"Syntax: sud0 [-v|-h] [-u uid] commandline" NL NL \
	"Options:" NL \
	" -v      print version and exit 1" NL \
	" -h      print this help screen and exit 1" NL \
	" -u uid  attempt to run commandline as user uid (integer)" NL NL \
	"sud0 reads " CFGFILE " to determine if running commandline is permitted." NL NL

uid_t my_uid, my_euid;
gid_t my_gid, my_egid;
int have_uid, have_euid, have_gid, have_egid;

#if defined(CFGSRC) != defined(CFGGEN)
#error CFGSRC and CFGGEN needs to be both or neither defined.
#endif

static void check_regen_conf()
{
	struct stat st_dst;
	if (stat(CFGFILE, &st_dst) != 0) {
		fprintf(stderr, "sud0: config file '%s' doesn't exist, create it first\n", CFGFILE);
		exit(1);
	}

#ifdef CFGUID
	if (st_dst.st_uid != CFGUID) {
		fprintf(stderr, "sud0: config file '%s' has the wrong owner. Need %ld, file is %ld\n", CFGFILE, (long)CFGUID, (long)st_dst.st_uid);
		exit(1);
	}
#endif

	if ((st_dst.st_mode & 077) != 0) {
		fprintf(stderr, "sud0: config file '%s' wrong permission: should not be readable or writable by group or others.\n", CFGFILE);
		exit(1);
	}

	/* call CFGGEN to re-generate the config if needed */
#ifdef CFGSRC
	{
		struct stat st_src;
		if (stat(CFGSRC, &st_src) != 0) {
			fprintf(stderr, "sud0: config source '%s' doesn't exist, create it first\n", CFGSRC);
			exit(1);
		}
		if (st_src.st_mtime > st_dst.st_mtime) {
			/* It seems, we must call setreuid() to get the generator script
			   work properly. This call might have side-effects, so it may
			   turn out this is not the right place to call this. */
			setreuid(0, 0);
			system(CFGGEN);
		}
	}
#endif
}

/* Decimal strtol that returns 1 (failure) with an error message on invalid input */
#define STRTOL(dst, src) \
do { \
	char *end; \
	dst = strtol(src, &end, 10); \
	if (*end != '\0') { \
		fprintf(stderr, "sud0: invalid condition integer value: '%s'\n", src); \
		return 1; \
	} \
} while(0)

/* Integer check if src == saved. If we don't yet have a saved value,
   use getter to save one */
#define CHECK_INT(src, saved, have_saved, getter) \
do { \
	long vl; \
	if (!have_saved) { \
		saved = getter; \
		have_saved = 1; \
	} \
	STRTOL(vl, src); \
	return (vl != saved); \
} while(0)

/* Quick strcmp for short strings */
#define STRCMP2(a, b) ((a[0] == b[0]) && (a[1] == b[1]) && (a[2] == '\0'))
#define STRCMP3(a, b) ((a[0] == b[0]) && (a[1] == b[1]) && (a[2] == b[2]) && (a[3] == '\0'))
#define STRCMP4(a, b) ((a[0] == b[0]) && (a[1] == b[1]) && (a[2] == b[2]) && (a[3] == b[3]) && (a[4] == '\0'))

static int condition_fails(char *cond, int target_uid)
{
	char *key = cond, *val = strchr(cond, '=');

	if ((val == NULL) || (val[1] != '='))
		return 1;

	*val = '\0';
	val += 2;

	if (STRCMP3(key, "uid"))  CHECK_INT(val, my_uid, have_uid, getuid());
	if (STRCMP3(key, "gid"))  CHECK_INT(val, my_gid, have_gid, getgid());
	if (STRCMP4(key, "euid")) CHECK_INT(val, my_euid, have_euid, getuid());
	if (STRCMP4(key, "egid")) CHECK_INT(val, my_egid, have_egid, getgid());
	if (STRCMP2(key, "-u")) {
		long vl;
		STRTOL(vl, val);
		return (vl != target_uid);
	}

	fprintf(stderr, "sud0: invalid condition key: '%s'\n", key);
	return 1;
}

/* Decide if a given command line is permitted by the config; returns 1
   on permit, 0 on deny */
static int permit_cmd(int argc, char *argv[], int target_uid)
{
	FILE *f;
	char *line, tmp[1024];

	if ((argc < 1) || (*argv[0] != '/')) {
		fprintf(stderr, "sud0: command to execute must have absolute path, starting with a /\n");
		return 0;
	}

	f = fopen(CFGFILE, "r");
	if (f == NULL) {
		fprintf(stderr, "sud0: can't open config file %s for read\n", CFGFILE);
		return 0;
	}

	/* read the config file line by line... */
	while((line = fgets(tmp, sizeof(tmp)-1, f)) != NULL) {
		int len, n, cargc, fail, cmd_idx;
		char **cargv, *end;

		tmp[sizeof(tmp)-1] = '\0'; /* don't trust fgets() corner case on random systems in a suid root executable! */

		/* ... skip empty lines and comments */
		while(isspace(*line)) line++;
		if ((*line == '\0') || (*line == '#')) continue;

		/* require a \n and remove it; missing \n normally means our tmp[] was too small */
		len = strlen(tmp);
		end = tmp + len - 1;
		if (*end != '\n') {
			fprintf(stderr, "sud0: broken config file (line too long?): '%s'\n", tmp);
			return 0;
		}
		*end = '\0';

		/* ... parse the line into cargc/cargv[] for the comparison */
		cargc = qparse2(line, &cargv, QPARSE_MULTISEP);

		/* check conditions */
		fail = 0;
		cmd_idx = 0;
		while((cmd_idx < cargc) && (cargv[cmd_idx][0] != '/')) {
			if (condition_fails(cargv[cmd_idx], target_uid)) {
				fail = 1;
				break;
			}
			cmd_idx++;
		}

		if (fail)
			continue;

		/* compare each user supplied arg to the args in the config file line */
		if ((cargc-cmd_idx) != argc) {
			qparse_free(cargc, &cargv);
			continue;
		}

		/* check command line */
		fail = 0;
		for(n = cmd_idx; n < argc; n++) {
			if ((cargv[n][0] == '*') && (cargv[n][1] == '\0')) continue; /* accept anything for * */
			if (strcmp(cargv[n], argv[n]) != 0) {

				fail = 1;
				break;
			}
		}
		qparse_free(cargc, &cargv);

		/* all matched - permit! */
		if (!fail) {
			fclose(f);
			return 1;
		}
	}

	/* finished with the config file, no match */
	fclose(f);
	return 0;
}

int main(int argc, char *argv[])
{
	char *env0[] = { NULL };
	int uid = 0;

	if ((argv[1] != NULL) && (argv[1][0] == '-') && (argv[1][1] == 'v') && (argv[1][2] == '\0')) {
		puts("sud0 version " SUD0_VER);
		puts("Web: http://www.repo.hu/projects/sud0");
		exit(1);
	}


	if (((argv[1] != NULL) && (argv[1][0] == '-') && (argv[1][1] == 'h') && (argv[1][2] == '\0')) || (argc == 1)) {
		puts(HELP);
		exit(1);
	}

	check_regen_conf();

	/* parse -u UID */
	if ((argv[1] != NULL) && (argv[1][0] == '-') && (argv[1][1] == 'u') && (argv[1][2] == '\0')) {
		long l;
		char *end;
		if (argv[2] == NULL) {
			fprintf(stderr, "sud0: -u requires an user ID\n");
			return 1;
		}
		l = strtol(argv[2], &end, 10);
		if (*end != '\0') {
			fprintf(stderr, "sud0: -u requires an integer user IDs\n");
			return 1;
		}
		if ((l < 0) || (l > 65535)) {
			fprintf(stderr, "sud0: -u user ID out of range: %ld\n", l);
			return 1;
		}
		uid = l;
		argc-=2;
		argv+=2;
	}

	/* the actual implementation */

	if (!permit_cmd(argc-1, argv+1, uid)) {
		fprintf(stderr, "sud0: permission denied (by config file)\n");
		return 1;
	}

	if (setuid(uid) != 0) {
		fprintf(stderr, "sud0: failed to setuid to %d\n", uid);
		return 1;
	}
	execve(argv[1], argv+1, env0);
	perror("sud0: exec failed:");
	return 1;
}
